/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package cricketdaemon.beaconstat;

import cricketdaemon.CricketDaemon;
import cricketdaemon.UnitConv;
import cricketdaemon.Util;
import cricketdaemon.module.*;
import cricketdaemon.reader.parse.ParseMap;
import cricketdaemon.reader.*;
import cricketdaemon.reader.sim.*;
import java.util.*;

/**
 * A module that keeps track of statistics of all the beacons heard
 * within a specified window period.
 *
 * @author Allen Miu
 */
public class BeaconStatModule extends PeriodicModuleThread 
    implements Beacon.BeaconHandler
{
    protected static BitSet mask;

    /* field types */
    public final static int STAT = 1;

    /* parameters */
    final static long REFRESH_RATE = 100;  // in milliseconds
    //final static long REFRESH_RATE = 1000;  // in milliseconds
    protected static long window; // in milliseconds
    protected int binSize;

    /* specifies the distanceEstimate to use mean, mode, med, max, min */
    protected int statSelect;
    protected int enoughSample;

    protected Hashtable beaconsHeard;

    protected long lastUpdate = 0;

    public BeaconStatModule()
    {
	beaconsHeard = new Hashtable();
	mask = new BitSet();
	mask.set(STAT);
    }

    public BeaconStatModule(ModuleThread reader, int w, int es, int ss)
    {
	super("BeaconStatModule", REFRESH_RATE);

	/* registering with the "port" reader */
	BitSet desiredTypeMask = new BitSet();
	desiredTypeMask.set(ParseMap.SPACE);
	desiredTypeMask.set(ParseMap.BEACONID);
	desiredTypeMask.set(ParseMap.DISTANCE);
	desiredTypeMask.set(ParseMap.LISTENTIME);
	reader.register(this, desiredTypeMask);

	/* reset */
	beaconsHeard = new Hashtable();
	mask = new BitSet();
	mask.set(STAT);
	window = w;
	statSelect = ss;
	enoughSample = es;
	binSize = 1;
    }

    //protected void periodicallyProcess()
    synchronized protected void periodicallyProcess()
    {
	long curtime = curTime();
	if((curtime - lastUpdate) >= REFRESH_RATE) {
	    Enumeration beacons = beaconsHeard.elements();
	    while(beacons.hasMoreElements()) {
		BeaconDistStat r = (BeaconDistStat) beacons.nextElement();
		r.refresh(curtime);
		if(r.shouldRemove()) {
		    beaconsHeard.remove(r.key);
		}
	    }
	    BeaconStatData result = beaconList();

	    //long compute = curTime() - curtime;

	    if(result != null) {
		//System.out.println(System.currentTimeMillis()+" BeaconStatModule sent");
		transfer(result, mask);

	    }
	    //else {
	    //System.out.println(System.currentTimeMillis()+" BeaconStatModule result==null");
	    //}

	    //long xfer = curTime() - curtime;
	    
	    lastUpdate = curtime;
	    //System.out.println("Sent!!!!!!! "+compute+" "+xfer);
	}
    }

    protected long curTime()
    {
	return System.currentTimeMillis();
    }

    public void processBeacon(Beacon newData, BitSet fieldMask)
    {
	//System.out.println("BeaconStatModule "+fieldMask);

	if(!fieldMask.get(ParseMap.SPACE)) {
	    Util.log("BeaconStatModule::processBeacon() space is null");
	    return; // silently drops
	}
	if(!fieldMask.get(ParseMap.DISTANCE)) {
	    Util.log("BeaconStatModule::processBeacon() distance is null");
	    return; // silently drops
	}

	// Object types are defined in reader.parse.ParseMap
	String space      = (String)  newData.getField(ParseMap.SPACE);
	Integer distance  = (Integer) newData.getField(ParseMap.DISTANCE);

	if(CricketDaemon.listenerVersion == CricketDaemon.CRICKETV1) {
	    if(distance.intValue() < UnitConv.BEACON_OFFSET_BODHI) {
		Util.log("BeaconStatModule::processBeacon() dist < offset ("+
			 distance.intValue()+" < "+UnitConv.BEACON_OFFSET_BODHI);
		return; 
		// drop the measurement because it is less than offset at zero
		// distance, which is impossible.
	    }

	    // subtract the offset to get the absolute time of flight value
	    distance = new Integer(distance.intValue() - UnitConv.BEACON_OFFSET_BODHI);
	}
	//else if(CricketDaemon.listenerVersion >= CricketDaemon.CRICKETV1_1) {
	    // "checksumming" the space id
	    // treat all space names that does not begin with "MIT" as errors
	    //System.out.println("Got Beacon "+CricketDaemon.listenerVersion+":"+
	    //space+"|"+distance+"|"+ (int)space.charAt(4)+"|"+
	    //space.length());
	    //System.out.println("Space:"+space+" "+space.startsWith("MIT-", 0));
	    
	    /*
	    if(! (space.startsWith("MIT-", 0) &&
		  (space.length() == 5) &&
		  (space.charAt(4) >= 48) &&
		  (space.charAt(4) <= 57))) {
		return;
	    }
	    */
	//}

	//System.out.println("BeaconStatModule "+distance+" "+CricketDaemon.MAXDIST);


	if(distance.intValue() > CricketDaemon.MAXDIST) {
	    // reject readings greater than threshold
	    return;
	}


	//System.out.println("Got Beacon "+CricketDaemon.listenerVersion+": "+
	//space+" "+distance+" "+fieldMask);

	insert(newData, fieldMask, distance);
    }

    protected void moduleCleanUp() { return; }

    /**
     * Inserts a new distance value read from the PortReader.
     */
    private synchronized void insert(Beacon beacon, BitSet fieldMask, Integer dist)
    {
	//System.out.println(System.currentTimeMillis()+" "+beacon.timeStamp+" BeaconStatModule recv");

	String space = (String)  beacon.getField(ParseMap.SPACE);
	Integer id   = null;
	if(fieldMask.get(ParseMap.BEACONID))
	    id  = (Integer) beacon.getField(ParseMap.BEACONID);
	else
	    id = new Integer(0);
	String uniqueID = space+"-"+id;


	Integer duration  = (Integer) beacon.getField(ParseMap.DURATION);

	Long listenTime = null;
	if(fieldMask.get(ParseMap.LISTENTIME))
	    listenTime = (Long) beacon.getField(ParseMap.LISTENTIME);
	else
	    listenTime = beacon.timeStamp;

	//System.out.println(uniqueID+" "+beaconsHeard.size());

	BeaconDistStat r = (BeaconDistStat) beaconsHeard.get(uniqueID); 

	if(r != null) // if in table, just insert the reading
	    r.insert(beacon, space, id.intValue(), dist, duration, listenTime);
	else {
	    r = new BeaconDistStat(uniqueID, beacon, space, 
				   id.intValue(), dist, duration, listenTime);
	    beaconsHeard.put(uniqueID, r);
	}
	periodicallyProcess();
    }

    /**
     * Returns a list of BeaconsHeard.
     */
    private synchronized BeaconStatData beaconList()
    {
	//ArrayList result = new ArrayList();

	// run analysis for all the beacons
	Collection beacons = beaconsHeard.values();
	//Object[] array = new Object[beacons.size()];

	long maxTime = 0;
	BeaconDistStat latestStat = null;
	

	ArrayList beaconArray = new ArrayList();
	Iterator it  = beacons.iterator();
	//int j = 0;
	while(it.hasNext()) {
	    BeaconDistStat r = (BeaconDistStat) it.next();
	    BeaconDistStat a = r.cloneAndAnalyze(statSelect, enoughSample, binSize);
	    
	    if(a != null) {
		beaconArray.add(a);
		if(a.beacon.hwTimeStamp.longValue() > maxTime) {
		    maxTime = a.beacon.hwTimeStamp.longValue();
		    latestStat = a;
		}
	    }
	    //	    System.out.println(a);
	}
	
	/*
	Object[] array = beaconArray.toArray();
	if (array.length <= 0)
	    return null;

	for(int i = 0; i < array.length; i++)
	    result.add(array[i]);
	*/


	// FIXME: beaconArray should be simply an array
	if(beaconArray.size() > 0)
	    return new BeaconStatData(latestStat.beacon, beaconArray, latestStat);
	else
	    return null;
    } 

    public static void main(String[] args)
    {
	ModuleThread r = new SimReader("sample.conf", null);
	ModuleThread b = new BeaconStatModule(r, 5000, 0, 0);
	ModuleThread d = new DebugModule(b);
	
	d.start();
	b.start();
	r.start();
	
	try {
	    r.join();
	}
	catch(InterruptedException e) {

	}
    }
}

