/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package cricketdaemon.beaconstat;

import java.util.*;

/**
 * An instantiated object of this class contains various statistical
 * data for a given input array of integers.
 *
 * @author Allen Miu
 */
public class Stat {
    
    public int binSize;

    public float median;
    public float mean;
    public float mode;
    public int max;
    public int min;
    public float stddev;
    public float variance;

    public Stat(int[] a, int bs)
    {
	binSize = bs;
	preprocessSample(a);

	// better to do a running avg instead
	median = median(a);
	mean   = mean(a);
	mode   = mode(a);
	max    = max(a);
	min    = min(a);
	variance = var(a);
	stddev = (float) Math.sqrt(variance);
    }

    // run this before invoking any other function
    private void preprocessSample(int[] a)
    {
	Arrays.sort(a);
    }
    
    private float sqr(float x) { return x*x; }
    
    private float var(int[] a)
    {
	if(a.length <= 1)
	    return 0;
	
	float mean = mean(a);
	float sum  = 0;
	for(int i = 0; i < a.length; i++)
	    sum += sqr(a[i] - mean);
	
	return sum/(a.length - 1);
    }

    private float mean(int[] a)
    {
	float sum = 0;
	for(int i = 0; i < a.length; i++)
	    sum += a[i];
	return sum/a.length;
    }

    private float mode(int[] a)
    {
	int prevDist = a[0] / binSize;
	prevDist *= binSize;

	//	int freq = 0;
	ArrayList modeDists = new ArrayList();
	ArrayList tempDists = new ArrayList();
	int modeFreq = 0;

	for(int i = 0; i < a.length; i++) {
	    int dist  = a[i] / binSize;
	    dist *= binSize;
	    
	    if(prevDist != dist) {
		if(tempDists.size() > modeFreq) {
		    modeDists.clear();
		    modeFreq = tempDists.size();
		}
		if(tempDists.size() >= modeFreq) {
		    for(int j = 0; j < tempDists.size(); j++)
			modeDists.add(tempDists.get(j));
		}
		tempDists.clear();
		prevDist = dist;
	    }

	    tempDists.add(new Integer(a[i]));

	    /*
	      if(prevDist == dist)
	      freq++;
	      else {
	      if(freq > modeFreq) {
	      modeDists.clear();
	      modeFreq = freq;
	      }
	      if(freq >= modeFreq)
	      modeDists.add(new Integer(prevDist));
	      prevDist = dist;
	      freq = 1;
	      }
	    */
	}
	
	if(tempDists.size() >= 1 || a.length == 1) { 
	    //	    if(freq > 1 || a.length == 1) { 
	    // run this block only if the "else" clause did not get run
	    // in the for-loop
	    if(tempDists.size() > modeFreq) {
		modeDists.clear();
		modeFreq = tempDists.size();
	    }
	    if(tempDists.size() >= modeFreq) {
		for(int i = 0; i < tempDists.size(); i++)
		    modeDists.add(tempDists.get(i));
	    }
	}
	int sum = 0;
	for(int i = 0; i < modeDists.size(); i++)
	    sum += ((Integer) modeDists.get(i)).intValue();
	
	return (float) sum/modeDists.size();
    }
		
    private int max(int[] a)
    {
	return a[a.length-1];
    }

    private int min(int[] a)
    {
	return a[0];
    }

    private float median(int[] a)
    {
	int middle = a.length/2;
	if(a.length % 2 == 0)
	    return (a[middle] + a[middle-1])/(float)2.0;
	else
	    return (float) a[middle];
    }

    public String toString()
    {
	return 
	    "med="+median+" avg="+mean+" mode="+mode+"\n"+
	    "max="+max+" min="+min+" stdd="+stddev+" var="+variance+"\n";
    }

    public static void main(String[] args)
    {
	int[] test1 = { 1, 1, 2, 3, 4, 4, 5, 6, 6, 6, 7 };
	int[] test2 = { 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10 };
	int[] test3 = { 1, 2, 3, 4, 4, 5, 6, 6, 6, 6, 6, 6, 6, 6 };
	int[] test4 = { 1, 1, 2, 3, 4, 4 };
	int[] test5 = { 3 };
	int[] test6 = { 1, 3, 3, 3, 3 };

	Stat s1 = new Stat(test1, 5);
	Stat s2 = new Stat(test2, 5);
	Stat s3 = new Stat(test3, 5);
	Stat s4 = new Stat(test4, 5);
	Stat s5 = new Stat(test5, 1);
	Stat s6 = new Stat(test6, 1);

	System.out.println(s1);
	System.out.println(s2);
	System.out.println(s3);
	System.out.println(s4);
	System.out.println(s5);
	System.out.println(s6);
	
    }
}
